/*
 * Copyright (C) 2008 OpenedHand Ltd
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * Initially based on http://simon.incutio.com/code/js/date-parser/dateparse.js,
 * ported to C by Matthew Allum <mallum@openedhand.com>.
 */

#include <config.h>
#define _XOPEN_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <glib.h>
#include <glib/gi18n.h>
#include "koto-date-parser.h"

typedef GDate* (*MatchFunc) (const char *input, GMatchInfo *info);

typedef struct Matcher {
  GRegex *regex;
  MatchFunc func;
} Matcher;

static GDate *
today (void)
{
  GDate *date;
  date = g_date_new ();
  /* Enforce a fixed today date when testing for ease */
#if TEST
  g_date_set_dmy (date, 11, 3, 2008);
#else
  g_date_set_time_t (date, time (NULL)); 
#endif
  return date;
}

static GDateWeekday
convert_str_weekday (const char *day)
{
  struct tm tm = {0,};
  char *p;
  p = strptime (day, "%a", &tm);
  if (p)
    /* God damn strptime/GDate impedance */
    if (tm.tm_wday == 0)
      return G_DATE_SUNDAY;
    else
      return tm.tm_wday;
  else
    return 0;
}

#if 0
static GDateMonth
convert_str_month (const char *month)
{
  struct tm tm = {0,};
  char *p;
  p = strptime (month, "%b", &tm);
  if (p)
    return tm.tm_mon + 1;
  else
    return 0;
}


static gboolean
match_day (const char  *input, 
	   size_t       n_matches,
	   gchar      **matches,
	   GDate       *date)
{
  g_date_set_time_t (date, time (NULL)); 
  g_date_set_day (date, atoi(matches[0]));

  return (g_date_valid (date));
}

static gboolean
match_day_month (const char  *input, 
		 size_t       n_matches,
		 gchar      **matches,
		 GDate       *date)
{
  g_date_set_time_t (date, time (NULL)); 
  g_date_set_day (date, atoi(matches[0]));
  g_date_set_month (date, convert_str_month(matches[1]));

  return (g_date_valid (date));
}

static gboolean
match_day_month_year (const char  *input, 
		      size_t       n_matches,
		      gchar      **matches,
		      GDate       *date)
{
  g_date_set_time_t (date, time (NULL)); 
  g_date_set_day (date, atoi(matches[0]));
  g_date_set_month (date, convert_str_month(matches[1]));
  g_date_set_year (date, atoi(matches[2]));

  return (g_date_valid (date));
}

static gboolean
match_month_day (const char  *input, 
		 size_t       n_matches,
		 gchar      **matches,
		 GDate       *date)
{
  g_date_set_time_t (date, time (NULL)); 
  g_date_set_day (date, atoi(matches[1]));
  g_date_set_month (date, convert_str_month(matches[0]));

  return (g_date_valid (date));
}

static gboolean
match_month_day_year (const char  *input, 
		      size_t       n_matches,
		      gchar      **matches,
		      GDate       *date)
{
  g_date_set_time_t (date, time (NULL)); 
  g_date_set_day (date, atoi(matches[1]));
  g_date_set_month (date, convert_str_month(matches[0]));
  g_date_set_year (date, atoi(matches[2]));

  return (g_date_valid (date));
}

static gboolean
match_last (const char  *input, 
	    size_t       n_matches,
	    gchar      **matches,
	    GDate       *date)
{
  gint day, last_day, sub_days;

  if (n_matches < 1) 
    return FALSE;

  g_date_set_time_t (date, time (NULL)); 
  day      = g_date_get_weekday (date);
  last_day = convert_str_weekday (matches[0]); 
  /* TODO: handle convert_str_weekday failing */
  sub_days = 7 - last_day + day;
  
  g_date_subtract_days (date, sub_days);

  return (g_date_valid (date));
}


static gboolean
match_digits (const char  *input, 
	      size_t       n_matches,
	      gchar      **matches,
	      GDate       *date)
{
  struct tm tm = {0,};
  char *p;

  if (n_matches != 1)
    return FALSE;

  p = strptime (matches[0], "%x", &tm);
  if (p) {
    if (tm.tm_mday)
      g_date_set_day (date, tm.tm_mday);
    if (tm.tm_mon)
      g_date_set_month (date, tm.tm_mon + 1);
    if (tm.tm_year)
      g_date_set_year (date, tm.tm_year + 1900);
    return (g_date_valid (date));
  } else {
    return FALSE;
  }
}

#endif

static GDate *
match_today (const char *input, GMatchInfo *info)
{
  GDate *date;
  date = today ();
  return date;
}

static GDate *
match_tomorrow (const char *input, GMatchInfo *info)
{
  GDate *date;
  date = today ();
  g_date_add_days (date, 1);
  return date;
}

static GDate *
match_yesterday (const char *input, GMatchInfo *info)
{
  GDate *date;
  date = today ();
  g_date_subtract_days (date, 1);
  return date;
}

static GDate *
match_next (const char  *input, GMatchInfo *info)
{
  GDate *date;
  char *s;
  GDateDay current_day, next_day;
  int add_days;
  
  s = g_match_info_fetch_named (info, "day");
  next_day = convert_str_weekday (s);
  g_free (s);
  if (next_day == G_DATE_BAD_WEEKDAY)
    return NULL;
  
  date = today ();
  current_day = g_date_get_weekday (date);
  add_days = 7 + next_day - current_day;

  g_date_add_days (date, add_days);
  
  return date;
}

static GDate *
match_this (const char *input, GMatchInfo *info)
{
  GDate *date;
  char *s;
  gint target_day, today_day;

  s = g_match_info_fetch_named (info, "day");
  target_day = convert_str_weekday (s);
  g_free (s);
  if (target_day == G_DATE_BAD_WEEKDAY)
    return NULL;
    
  date = today ();
  today_day = g_date_get_weekday (date);
  g_date_subtract_days (date, today_day);
  g_date_add_days (date, target_day);

  return date;
}

static GDate *
match_date (const char *input, GMatchInfo *info)
{
  GDate *date;
  char *day_s, *month_s;
  GDateDay day;
  GDateMonth month;

  day_s = g_match_info_fetch_named (info, "day");
  if (!day_s)
    return NULL;
  day = atoi (day_s);
  g_free (day_s);

  month_s = g_match_info_fetch_named (info, "month");
  if (!month_s)
    return NULL;
  month = atoi (month_s);
  g_free (month_s);
  
  if (!g_date_valid_day (day) || !g_date_valid_month (month))
    return NULL;

  date = today ();
  g_date_set_day (date, day);
  g_date_set_month (date, month);
  
  return date;
}


static Matcher*
matcher_new (const gchar *regex_str, MatchFunc func)
{
  GError *error = NULL;
  GRegex *regex;
  Matcher *m;

  g_assert (regex_str);
  g_assert (func);
  
  regex = g_regex_new (regex_str,
		       G_REGEX_CASELESS | G_REGEX_OPTIMIZE, 0,
		       &error);
  
  if (error) {
    g_warning ("Cannot parse regex '%s': %s", regex_str, error->message);
    g_error_free (error);
    return NULL;
  }

  m = g_slice_new (Matcher);
  m->regex = regex;
  m->func = func;  
  return m;
}

static GList *
init_matchers (void)
{
  GList *matchers = NULL;
  Matcher *m;
  
  /* Translators: just translate "today" */
  m = matcher_new (_("^(?<task>.+) today$"), match_today);
  if (m) matchers = g_list_prepend (matchers, m);

  /* Translators: just translate "tomorrow" */
  m = matcher_new (_("^(?<task>.+) tomorrow$"), match_tomorrow);
  if (m) matchers = g_list_prepend (matchers, m);

  /* Translators: just translate "yesterday" */
  m = matcher_new (_("^(?<task>.+) yesterday$"), match_yesterday);
  if (m) matchers = g_list_prepend (matchers, m);

  /* Translators: just translate "next" */
  m = matcher_new (_("^(?<task>.+) next (?<day>\\w+)$"), match_next);
  if (m) matchers = g_list_prepend (matchers, m);

  /* Translators: just translate "this" */
  m = matcher_new (_("^(?<task>.+) this (?<day>\\w+)$"),  match_this);
  if (m) matchers = g_list_prepend (matchers, m);

  /* Translators: replace by|due|on with suitable words, and re-order the day
     and month expressions as suitable */
  m = matcher_new (_("^(?<task>.+) (?:by|due|on) (?<day>\\d{1,2})/(?<month>\\d{1,2})$"),  match_date);
  if (m) matchers = g_list_prepend (matchers, m);

#if 0
  m = matcher_new ("^(\\d{1,2})(st|nd|rd|th)?$", match_day);
  if (m) matchers = g_list_prepend (matchers, m);

  m = matcher_new ("^(\\d{1,2})(?:st|nd|rd|th)? (\\w+)$", match_day_month);
  if (m) matchers = g_list_prepend (matchers, m);

  m = matcher_new ("^(\\d{1,2})(?:st|nd|rd|th)? (\\w+),? (\\d{4})$", 
		  match_day_month_year);
  if (m) matchers = g_list_prepend (matchers, m);

  m = matcher_new ("^(\\w+) (\\d{1,2})(?:st|nd|rd|th)?$", match_month_day);
  if (m) matchers = g_list_prepend (matchers, m);

  m = matcher_new ("^(\\w+) (\\d{1,2})(?:st|nd|rd|th)?,? (\\d{4})$", 
		  match_month_day_year);
  if (m) matchers = g_list_prepend (matchers, m);

  m = matcher_new ("^last (\\w+)$",  match_last);
  if (m) matchers = g_list_prepend (matchers, m);

  m = matcher_new ("^([\\d/-]+)$", match_digits);
  if (m) matchers = g_list_prepend (matchers, m);
#endif

  return g_list_reverse (matchers);
}

GDate *
koto_date_parse (const char *s, char **out_s)
{
  static GList *matchers = NULL;
  GList  *l;
  GDate *date = NULL;

  if (matchers == NULL)
    matchers = init_matchers ();
  
  for (l = matchers; l != NULL; l = l->next) {
    Matcher *m = l->data;
    GMatchInfo *info;

    if (g_regex_match (m->regex, s, 0, &info)) {
      date = m->func (s, info);
      if (!date)
	continue;
      
      if (!g_date_valid (date)) {
	g_date_free (date);
	continue;
      }
      
      if (out_s)
	*out_s = g_match_info_fetch_named (info, "task");
      break;
    }
    g_match_info_free (info);
  }

  return date;
}

#if TEST

#include <locale.h>
#include <string.h>

static void
compare (GDate *expected, GDate *date)
{
  g_assert (g_date_valid (date));
  if (g_date_compare (date, expected) != 0) {
    char s1[32], s2[32];
    g_date_strftime (s1, sizeof (s1), "%x", expected);
    g_date_strftime (s2, sizeof (s2), "%x", date);
    g_printerr ("Incorrect date: expected %s got %s\n", s1, s2);
    g_assert_not_reached ();
  }
}

int
main (int argc, char **argv)
{
  GDate *test, *date;
  char *s = NULL;

  /* Set the locale to C so we can do string comparisons easily */
  setlocale(LC_ALL, "C");

  /* Today, tomorrow, yesterday */

  test = today ();
  date = koto_date_parse ("foo today", &s);
  g_assert (strcmp (s, "foo") == 0); g_free (s);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo tomorrow", &s);
  g_assert (strcmp (s, "foo") == 0); g_free (s);
  compare (test, date);
  g_date_free (date);

  g_date_subtract_days (test, 2);
  date = koto_date_parse ("foo yesterday", &s);
  g_assert (strcmp (s, "foo") == 0); g_free (s);
  compare (test, date);
  g_date_free (date);
  
  /* This [weekday] */

  g_date_set_dmy (test, 10, 3, 2008);
  date = koto_date_parse ("foo this monday", &s);
  g_assert (strcmp (s, "foo") == 0); g_free (s);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo this tuesday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo this wednesday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo this thursday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo this friday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo this saturday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo this sunday", NULL);
  compare (test, date);
  g_date_free (date);

  /* Next [weekday] */

  g_date_set_dmy (test, 17, 3, 2008);
  date = koto_date_parse ("foo next monday", &s);
  g_assert (strcmp (s, "foo") == 0); g_free (s);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo next tuesday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo next wednesday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo next thursday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo next friday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo next saturday", NULL);
  compare (test, date);
  g_date_free (date);

  g_date_add_days (test, 1);
  date = koto_date_parse ("foo next sunday", NULL);
  compare (test, date);
  g_date_free (date);

  return 0;
}
#endif
